const coral = @import("coral");

const ext = @cImport({
	@cInclude("spirv-cross/spirv_cross_c.h");
});

const sokol = @import("sokol");

const std = @import("std");

pub const CompileError = std.mem.Allocator.Error || error {
	UnsupportedTarget,
	InvalidSpirV,
	UnsupportedSpirv,
	UnknownFailure,
};

pub const Sources = struct {
	vertex_spirv: []const u32,
	fragment_spirv: []const u32,
};

pub fn compile(arena: *std.heap.ArenaAllocator, sources: Sources) CompileError!sokol.gfx.ShaderDesc {
	var context = @as(ext.spvc_context, null);

	try switch (ext.spvc_context_create(&context)) {
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	ext.spvc_context_set_error_callback(context, log_errors, null);

	defer ext.spvc_context_destroy(context);

	const arena_allocator = arena.allocator();
	const shader_source_sentinel = @as(u8, 0);

	return .{
		.vs = .{
			.source = try arena_allocator.dupeZ(u8, coral.io.slice_sentineled(
				shader_source_sentinel,
				@ptrCast(try compile_shader(context, ext.SpvExecutionModelVertex, sources.vertex_spirv)))),
		},

		.fs = .{
			.source = try arena_allocator.dupeZ(u8, coral.io.slice_sentineled(
				shader_source_sentinel,
				@ptrCast(try compile_shader(context, ext.SpvExecutionModelFragment, sources.fragment_spirv)))),
		},
	};
}

fn compile_shader(context: ext.spvc_context, model: ext.SpvExecutionModel, spirv: []const u32) CompileError![*]const u8 {
	var parsed_ir = @as(ext.spvc_parsed_ir, null);

	try switch (ext.spvc_context_parse_spirv(context, spirv.ptr, spirv.len, &parsed_ir)) {
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_INVALID_SPIRV => error.InvalidSpirV,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	var compiler = @as(ext.spvc_compiler, null);

	try switch (ext.spvc_context_create_compiler(context, ext.SPVC_BACKEND_GLSL, parsed_ir, ext.SPVC_CAPTURE_MODE_TAKE_OWNERSHIP, &compiler)) {
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_INVALID_ARGUMENT => error.UnsupportedTarget,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	try switch (ext.spvc_compiler_build_combined_image_samplers(compiler)) {
		ext.SPVC_ERROR_UNSUPPORTED_SPIRV => error.UnsupportedSpirv,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	try switch (ext.spvc_compiler_set_entry_point(compiler, "main", model)) {
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	var options = @as(ext.spvc_compiler_options, null);

	try switch (ext.spvc_compiler_create_compiler_options(compiler, &options)) {
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	try switch (ext.spvc_compiler_options_set_uint(options, ext.SPVC_COMPILER_OPTION_GLSL_VERSION, 430)) {
		ext.SPVC_ERROR_INVALID_ARGUMENT => error.InvalidSpirV,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	try switch (ext.spvc_compiler_options_set_bool(options, ext.SPVC_COMPILER_OPTION_GLSL_ES, @intFromBool(false))) {
		ext.SPVC_ERROR_INVALID_ARGUMENT => error.InvalidSpirV,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	try switch (ext.spvc_compiler_options_set_bool(options, ext.SPVC_COMPILER_OPTION_GLSL_VULKAN_SEMANTICS, @intFromBool(false))) {
		ext.SPVC_ERROR_INVALID_ARGUMENT => error.InvalidSpirV,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	try switch (ext.spvc_compiler_install_compiler_options(compiler, options)) {
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	var source = @as([*]const u8, undefined);

	try switch (ext.spvc_compiler_compile(compiler, @ptrCast(&source))) {
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_UNSUPPORTED_SPIRV => error.UnsupportedSpirv,
		ext.SPVC_SUCCESS => {},
		else => error.UnknownFailure,
	};

	return source;
}

fn log_errors(userdata: ?*anyopaque, error_message: [*c]const u8) callconv(.C) void {
	std.debug.assert(userdata == null);
	std.log.err("{s}", .{error_message});
}
